"""
Spectra
"""


import numpy as np
import pyfits as pf
import math
import glob
import lib
import copy as pythoncopy
import tem
import scipy.signal as spsig
import sys

class spectra:
    """a spectra is a flux assiociated to wavelengths, the other parameters can be readfrom the SDSS data """
	
    def __init__(self, fluxes_cs = np.array([]), fluxes=np.array([]), noise =np.array([]), zsdss = 0, coeff0 = 3.5793, coeff1 = 0.0001, ra = 0, dec = 0, angst =np.array([])):
	self.angst = angst #an array of wavelengths (in angstroms)
	self.fluxes = fluxes #flux (with continuum)
	self.fluxes_cs = fluxes_cs # flux (without continuum)
        self.fluxes_csc = np.array([]) #flux (without continuum, calculated)
        self.noise = noise
	self.zsdss = zsdss # redshift calculated by SDSS
        self.secz = 0
        self.mask = np.array([])
	self.plotcolour = "black"
        self.coeff0 = coeff0 # Center wavelength (log10) of first pixel
        self.coeff1 = coeff1 # Log10 dispersion per pixel
        self.ra = ra
        self.dec = dec
        self.name = 'not from fit file'
        self.allfluxes = {'fluxes' : self.fluxes, 'fluxes_cs' : self.fluxes_cs, 'fluxes_csc' : self.fluxes_csc}
        
    def initmask(self):
        """initialize the mask to an array containing only 'False' values, this fonction should be called only once the spectra is non-empty"""
        self.mask = np.ones(len(self.angst))
        self.mask = np.array(self.mask != 1)
        
    def copy(self):
        """Returns a deep copy of itself"""
        return pythoncopy.deepcopy(self)

    def pixtoangst(self, index):
        """converts a pixel to a wavelenght in angstroems, using a formula from SDSS site"""
        return 10**(self.coeff0 + self.coeff1 * index)
        
    def angsttopix(self, wavelength):
        """converts a wavelength in angstroems to a pixel (it is the inverse function 'pixtoangst')"""
        return math.floor((math.log(wavelength,10) - self.coeff0)/self.coeff1)

    def pixtoz(self, index):
        """converts a pixel to a redshift"""
        return 10**(self.coeff1*index)-1

    def ztopix(self, z):
        """converts a redshift to a pixel (it is the inverse function 'pixtoz')"""
        return int(math.log((z + 1), 10)/self.coeff1) 

    def setangst(self, coeff0 = 3.5793, coeff1 = 0.0001):
        """defines the wavelength of the spectra"""
        i =np.arange(1, len(self.fluxes_cs)+1) #individual pixel number array
        self.angst =np.array( map(self.pixtoangst,i) )
        
    def setz(self, z):
        """sets the wavelengths corrsponding to the redshift"""
        self.angst =(1 + z) / (1 + self.zsdss) * self.angst    

    def calc_csc(self):
        """substracts the continuum of the spectra, using a median filter"""
        kernel = 31
        smooth = spsig.medfilt(self.fluxes, kernel)
        self.fluxes_csc = self.fluxes - smooth
        self.allfluxes['fluxes_csc'] = self.fluxes_csc
        

    def read(self, directory, filename):
	"""creates a spectra from a SDSS file (i.e. reads the redshift, the flux, the noise...etc."""
	hdulist = pf.open(filename)
        name = filename[len(directory):len(filename)]
	fluxes = np.array(hdulist[0].data[0]) 
        fluxes_cs =np.array( hdulist[0].data[1])
        noise =np.array( hdulist[0].data[2])
        coeff0 = hdulist[0].header['COEFF0'] 
        coeff1 = hdulist[0].header['COEFF1'] 
        ra = hdulist[0].header['RA'] 
        dec = hdulist[0].header['DEC'] 
        zsdss = hdulist[0].header['Z'] 
        hdulist.close()
       
	sp = spectra(fluxes_cs, fluxes, noise,  zsdss, coeff0, coeff1, ra, dec)
        sp.name = name
        sp.setangst(coeff0, coeff1)
        sp.calc_csc()
        
	return sp


    def mask_angst(self, tupleslist, z = 0):
        """for a spectra, we give a list of tuples (each one containing a wavelength and a width) and we hide the pixels at the given wavelength"""
        if tupleslist != [] and tupleslist != None:
            for tuple in tupleslist:
                n =  self.angsttopix((1 + z) * tuple[0])
                i1 = self.angsttopix((1 + z) * tuple[0])-self.angsttopix((1 + z) * (tuple[0]-tuple[1]/2.0))
                i2 = self.angsttopix((1 + z) * (tuple[0]+tuple[1]/2.0)) - self.angsttopix((1 + z) * tuple[0])
                self.mask[n - i1:n + i2 + 1] = True 
        

    def mask_height(self, height, fluxchoice):
        """hides the peaks higher than a given value [not used]""" 
        a = self.allfluxes[fluxchoice] < height
        self.mask[a] = True 
        
    def mask_borders(self, lambda1 = 5007 + 20, lambda2 = 6562.81 - 160):
        """eliminates the flux outside the two given wavelengths"""
        if lambda2 < lambda1:
            x = lambda1
            lambda1 = lambda2
            lambda2 = x

        lambda1 = lambda1 * (self.zsdss + 1)
        lambda2 = lambda2 * (self.zsdss + 1)
            
        n1 = self.angsttopix(lambda1)
        n2 = self.angsttopix(lambda2)
        size = len(self.fluxes_cs)
                                            
        if n1 > 0:
            self.mask[0:n1] = True
        if n2 < size:
            self.mask[n2:] = True
    

    def specsim(self, z1, z2, model, noiseintensity, secintensity = 1):
        """simulates a spectra containing one main  and one secondary  spectra : you give one spectra as a model, the noise intensity, the two redshift and the ratio between the two spectra flux [not used]"""
        self.zsdss = model.zsdss
        self.secz = z2
        self.fluxes_cs = model.fluxes_cs
        self.noise = model.noise
        self.setangst()
        self.setz(z1)
        self.zsdss = z1
        
        size = len(self.fluxes_cs)
        shifting = self.ztopix(z2)-self.ztopix(z1)
        b = np.zeros((1,size))[0]
        b[shifting:size-1] = self.fluxes_cs[0:size-shifting-1]
        b *= secintensity
        self.fluxes_cs = self.fluxes_cs + b
        noise = noiseintensity*(np.random.rand(size)-0.5)
        self.fluxes_cs = self.fluxes_cs + noise
        self.calc_csc()
 

class crosscorrspec:
    """This is a crosscorrelation spectra : it contains the origin spectra, a template which we use for the crosscorrelation and the result"""
    def __init__(self, spectra = spectra(), template = tem.template(), corr = np.array([])):
        self.spectra = spectra
        self.template = template
        self.corr = corr

    def crosscorr(self, fluxchoice):
        """makes the crosscorrelation between the spectra and the templates : it starts when both arrays are 'face to face', then it shifts pixel by pixel until the last one """        
        corr = []
        
        model = np.array(map(self.template.flux_gauss, self.spectra.angst))
        a = self.spectra.allfluxes[fluxchoice].copy()
        a[self.spectra.mask] = 0
        index1 = len(a)
        index2 = len(model)
        """
        #this part was made to avoid useless calculations, but with standards parameters, it didn't make it quicker, so I abandonned it. Maybe it can improved to be effective. 
        i = 0
        mod = model[0:index2-1-i]
        while np.any(mod > 0.01) :
            corr = np.append(corr, np.sum(self.spectra.fluxes_cs[i:index1-1]*mod))
            i += 1
            mod = model[0:index2-1-i]
        #print list(mod)
        #sys.exit()
        self.corr = corr
        """
        
        for i in range(index1):
            corr.append(np.sum(a[i:index1-1]*model[0:index2-1-i]))
        self.corr = np.array(corr)
        

def spfactory(directory):
    """factory function for a list of spectra : if you give the name of a directory, it will return a list of spectra obtained from every fits files in the directory [not used]"""

    spectralist = []
    
    filelist = glob.glob(directory + '*.fit')
    for file in filelist:
        sp = spectra()
        sp = sp.read(directory, file)
        
        spectralist.append(sp)
        
    return spectralist


def cspfactory(spectralist, template, fluxchoice):
    """factory function for a list of crosscorrelate spectra [not used]"""
    crosssplist = []
    for sp in spectralist:
        c = crosscorrspec(sp, template)
        c.crosscorr(fluxchoice)
        crosssplist.append(c)
    return crosssplist
